from . import sources
from .base import BaseChapter, BaseMetadata
from .sources.base_source import BaseSource


class BaseComic:
    """
    A comic with metadata and chapter data.

    :param `metadata`: Metadata of the comic
    :param `chapters`: A list of chapters of the comic
    """

    def __init__(
        self,
        metadata: BaseMetadata,
        chapters: list[BaseChapter],
    ):
        self.metadata = metadata
        self.chapters = chapters
        try:
            self.source = sources.get_class_for(self.metadata.url)(self.metadata.url)
        except ValueError as err:
            if self.metadata.url == "":  # sentinel value
                self.source = BaseSource("")
            else:
                raise ValueError from err

    def asdict(self) -> dict:
        """
        Return a dictionary representation of the comic.
        """

        return {
            "metadata": self.metadata.asdict(),
            "chapters": [c.asdict() for c in self.chapters],
        }

    def get_chapter_image_urls(self, chapter: BaseChapter) -> list[str]:
        """
        Fetch a list of image URLs of a chapter based on the
        current source.

        :param `chapter`: The chapter to fetch
        :return: A list of image URLs
        """
        return self.source.fetch_chapter_image_list(chapter)

    def set_chapter_range(self, *, start: int | None = None, end: int | None = None) -> None:
        """
        `start` and `end` are zero-indexed.

        :param `start`: The index of the first chapter to keep
        :param `end`: The index of the last chapter to keep (exclusive)
        """
        self.chapters = self.chapters[start:end]

    def update(self, *, chapters: bool = True, metadata: bool = True) -> None:
        """
        Refresh comic.metadata and comic.chapters with new information
        from the source. Remember to call mandown.save_metadata(comic)
        to actually save the updated data to the filesystem.

        :param `chapters`: whether to update the chapter index
        :param `metadata`: whether to update comic metadata
        """
        if chapters:
            self.chapters = self.source.fetch_chapter_list()

        if metadata:
            self.metadata = self.source.fetch_metadata()

    def __str__(self) -> str:
        return f"""
Title: {self.metadata.title},
Author(s): {', '.join(self.metadata.authors)}
URL: {self.metadata.url}
Genres: {', '.join(self.metadata.genres)}
Chapters: {len(self.chapters)}
Description:
    {self.metadata.description}
""".strip()



from enum import Enum
from pathlib import Path
from typing import Iterator

import comicon


class ConvertFormats(str, Enum):
    """
    The formats that mandown can convert to. This is used for the `--format` option.
    """

    # for typing purposes
    CBZ = "cbz"
    EPUB = "epub"
    PDF = "pdf"
    MOBI = "mobi"
    NONE = "none"


def convert_one(
    comic: comicon.Comic, comic_path: Path, to: ConvertFormats, dest_folder: Path
) -> Iterator[str | int]:
    # save comicon.json
    (comic_path / comicon.cirtools.IR_DATA_FILE).write_text(comic.to_json())

    yield from comicon.outputs.create_comic_progress(
        comic_path, dest_folder / f"{comic.metadata.title_slug}.{to.value}"
    )



# pylint: disable=invalid-name

"""
This module contains all the source modules.

The source modules are dynamically imported and added to the __class_list
variable. This variable is used by the get_class_for() function to return
the correct source class for a given URL.
"""

import sys
import types

from . import (
    source_batoto,
    source_blogtruyenmoi,
    source_comixextra,
    source_kuaikanmanhua,
    source_mangadex,
    source_mangakakalot,
    source_manganato,
    source_mangasee,
    source_manhuaes,
    source_readcomiconline,
    source_thecomicseries,
    source_webtoons,
)
from .common_source import CommonSource

__class_list: list[type[CommonSource]] = []


def _get_all_source_modules() -> list[str]:
    """
    Return a list of all source modules.
    """
    out = list[str]()
    for _, val in globals().items():
        if isinstance(val, types.ModuleType) and val.__name__.startswith("mandown.sources.source_"):
            out.append(val.__name__)
    return out


for i in _get_all_source_modules():
    __class_list.append(sys.modules[i].get_class())


def get_class_for(url: str) -> type[CommonSource]:
    """
    Return a source that matches the URL.

    :raises ValueError: If no source matches the URL.
    """
    for c in __class_list:
        if c.check_url(url):
            return c
    raise ValueError("No sources found matched the URL query.")


def get_all_classes() -> list[type[CommonSource]]:
    """
    Return all imported source module classes. Best used for manual poking.
    """
    return __class_list



"""
Source file for readmanganato.com
"""
# pylint: disable=invalid-name

import re
from urllib.parse import parse_qs, urlparse

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class MangaNatoSource(CommonSource):
    name = "ComicFury"
    domains = [
        "https://thecomicseries.com",
        "https://comicfury.com",
    ]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(
            requests.get(f"https://comicfury.com/comicprofile.php?url={self.id}").text,
            "lxml",
        )

        title = str(soup.select_one("div.authorname").text).strip()
        authors = [str(a.text).strip() for a in soup.select("a.authorname")]
        genres = [str(g.text).strip() for g in soup.select(".description-tags > a")]
        description_maybe = soup.select_one(".pccontent:has(.description-tags)")

        if description_maybe is not None:
            for div in description_maybe.find_all("div"):
                div.replace_with("\n")
        description = str(description_maybe.text).strip()
        cover = f"https://comicfury.com{str(soup.select_one('.profile-avatar > img')['src'])}"

        return BaseMetadata(title, authors, self.url, genres, description, cover)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(
            requests.get(f"https://comicfury.com/read/{self.id}/archive").text,
            "lxml",
        )

        chapters: list[BaseChapter] = [
            BaseChapter(e.text.strip(), f"https://comicfury.com{e['href']}")
            for e in soup.select("a:has(.archive-chapter)")
        ]
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        pages = soup.select(".archive-comics > a")
        # href is of form /read/title/comics/number
        first_page_id = pages[0]["href"].split("/")[-1]

        # /comic.php?action=addsubscription&amp;cid=number
        comic_id = soup.select_one(".webcomic-subscribe")["href"].split("=")[-1]

        # we need the total number of pages
        page_list_urls = soup.select("div.archive-pages .vfpage")
        index = None
        if page_list_urls:
            for i, el in enumerate(page_list_urls):
                if i == 0:
                    continue
                if "vfpagecurrent" in el["class"]:
                    # these appear twice so this is how we differentiate them
                    index = i - 1
                    break

        if index is None:
            num_pages = len(pages)
        else:
            soup3 = BeautifulSoup(
                requests.get(f"https://comicfury.com{page_list_urls[index]['href']}").text,
                "lxml",
            )
            # index is zero-indexed
            num_pages = len(pages) * (index + 1) + len(soup3.select(".archive-comics > a"))

        # their api only returns images after the first page
        # so we have to fetch it ourselves
        soup4 = BeautifulSoup(
            requests.get(f"https://comicfury.com{pages[0]['href']}").text,
            "lxml",
        )
        first_page = soup4.select_one(".is--comic-content img")["src"]

        # call their api
        page_id = first_page_id
        all_images: list[str] = [first_page]
        while len(all_images) < num_pages:
            data = requests.get(
                f"https://comicfury.com/api.php?url=webcomic/id/{comic_id}/comicid/{page_id}/getonsitereadercomics"
            ).json()
            if not data["status"] or data["error_code"]:
                raise RuntimeError(
                    "ComicFury did not give an expected response."
                    "Please report this issue to GitHub."
                )
            page_id = data["data"]["newLastComicId"]

            soup2 = BeautifulSoup(data["data"]["html"], "lxml")
            cur_images = [img["src"] for img in soup2.select(".is--comic-content img")]
            all_images.extend(cur_images)
            if data["data"]["endsAtLastComic"]:
                break
        return all_images

    @classmethod
    def url_to_id(cls, url: str) -> str:
        parsed = urlparse(url)
        if parsed.netloc == "comicfury.com":
            # it is in the form comicfury.com/comicprofile.php?url=name
            # or comicfury.com/read/*/comics/*
            if parsed.path.startswith("/read/"):
                return parsed.path.split("/")[2]
            return parse_qs(parsed.query)["url"][0]
        # it is in the form name.thecomicseries.com
        return parsed.netloc.split(".")[0]

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(r"https://comicfury\.com/comicprofile\.php.*", url)
            or re.match(r"https://.*\.thecomicseries\.com.*", url)
            or re.match(r"https://comicfury.com/read/.*/comics/*", url)
        )


def get_class() -> type[CommonSource]:
    return MangaNatoSource



import json
import multiprocessing as mp
import os
import urllib.parse
from pathlib import Path
from typing import Iterator, Sequence

import filetype
import requests as RealRequests
from natsort import natsorted

from .base import BaseChapter, BaseMetadata
from .comic import BaseComic

NUM_LEFT_PAD_DIGITS = 5
FILE_PADDING = f"0{NUM_LEFT_PAD_DIGITS}"
MD_METADATA_FILE = "md-metadata.json"

AsyncDownloadImageInput = tuple[str, Path | str, str | None, dict[str, str] | None]


def async_download_image(data: AsyncDownloadImageInput) -> None:
    """
    Download an image from a URL to a destination folder, fixing the file extension if necessary.

    :param `data`: A tuple of the url, destination folder, filename, and headers.
    """
    url, dest_folder, filename, headers = data
    dest_folder = Path(dest_folder)

    name = filename or url.split("/")[-1]
    dest_file = dest_folder / name

    res = RealRequests.get(url, headers=headers, timeout=5)

    if res.status_code != 200:
        # there is no clean way to raise an error in a pool
        # so we just return early and check it later
        return

    with open(dest_file, "wb") as file:
        file.write(res.content)

    # if the file extension is lying
    # rename it so epubcheck doesn't yell at us
    ext = filetype.guess(dest_file)
    if ext is not None and ext.extension in ["jpg", "png", "gif"]:
        dest_file.rename(dest_file.with_suffix(f".{ext.extension}"))


def download_images(
    urls: Sequence[str],
    dest_folder: Path | str,
    *,
    filestems: Sequence[str] | None = None,
    headers: dict[str, str] | None = None,
    threads: int = 1,
) -> Iterator[None]:
    """
    Download one or multiple URLs to a destination folder.
    Raises ValueError if the folder does not exist.

    :param `urls`: A list of URLs to download.
    :param `dest_folder`: The path to download files into.
    :param `filestems`: Specify the name of each downloaded file instead of the default.
    :param `headers`: Request headers
    :param `threads`: The number of processes to open
    :returns An Iterator that yields `None` for each downloaded file.
    """
    dest_folder = Path(dest_folder)

    # attempt to create
    dest_folder.mkdir(exist_ok=True)

    # args to async_download
    map_pool: list[AsyncDownloadImageInput] = []

    if filestems is None:
        filestems = [f"{i+1:FILE_PADDING}" for i in range(len(urls))]

    for url, stem in zip(urls, filestems, strict=True):
        _, ext = os.path.splitext(urllib.parse.urlparse(url).path)
        map_pool.append((url, dest_folder, f"{stem}{ext}", headers))

    with mp.Pool(threads) as pool:
        yield from pool.imap_unordered(async_download_image, map_pool)


def read_comic(path: Path | str) -> BaseComic:
    """
    Open a comic from a folder path.

    :param `path`: A folder containing `md-metadata.json`
    :returns A comic with metadata and chapter data of that folder
    :raises `FileNotFoundError` if `md-metadata.json` is not found.
    :raises `IOError` if the path does not exist.
    """
    path = Path(path)
    json_path = path / MD_METADATA_FILE

    if not path.exists():
        raise IOError(f"Path {path} does not exist")

    with open(json_path, "r", encoding="utf-8") as file:
        data = json.load(file)

    return BaseComic(
        BaseMetadata(**data["metadata"]),
        [BaseChapter(**c) for c in data["chapters"]],
    )


def parse_comic(path: Path | str, donor_comic: BaseComic | None = None) -> BaseComic:
    """
    Attempt to construct a Mandown comic from a folder from the filesystem (without most metadata).
    This function is non-destructive and does not modify the filesystem. If you want to save the
    comic, call `save_comic` on the returned `BaseComic`.

    :param `path`: A folder containing images
    :param `donor_comic`: A comic to fill metadata from
    :returns A `BaseComic` with metadata and chapter data of that folder
    """
    path = Path(path)

    title = path.stem
    authors: list[str] = []
    url = ""
    genres: list[str] = []
    description = ""
    cover_art = ""
    metadata = BaseMetadata(title, authors, url, genres, description, cover_art)

    chapters = [
        BaseChapter(inode.stem, "", inode.stem)
        for inode in natsorted(path.iterdir(), key=lambda i: i.stem)
        if inode.is_dir()
    ]

    if donor_comic:
        metadata = donor_comic.metadata
        for local, remote in zip(chapters, donor_comic.chapters):
            local.title = remote.title
            local.url = remote.url

    return BaseComic(metadata, chapters)


def save_comic(comic: BaseComic, path: Path | str) -> None:
    """
    Save an `md-metadata.json` from `comic` into `path`. If `path` does not exist, it will
    be created. THIS FUNCTION IS DESTRUCTIVE AND WILL OVERWRITE ANY EXISTING `md-metadata.json`.

    :param `comic`: A comic to save
    :param `path`: A folder to save `md-metadata.json` into
    """
    path = Path(path)
    path.mkdir(exist_ok=True)

    json_path = path / MD_METADATA_FILE

    with open(json_path, "w", encoding="utf-8") as file:
        json.dump(comic.asdict(), file)


def discover_local_images(path: Path | str) -> dict[str, list[Path]]:
    """
    Given a comic path, return a dictionary of chapter_slugs OR cover: images.
    Basically a slightly modified version of os.walk. Only recurses one level deep.

    :param `path`: A folder containing images
    :returns A dictionary of {slugs: images}
    """
    path = Path(path)

    return {
        chap.stem: sorted(chap.iterdir())
        for chap in sorted(path.iterdir())  # iterdir does not guarantee any order
        if chap.is_dir()
    } | {"cover": [cover for cover in path.iterdir() if cover.is_file() and cover.stem == "cover"]}



"""
Source file for readmanganato.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class MangaNatoSource(CommonSource):
    name = "MangaNato"
    domains = [
        "https://manganato.com",
        "https://readmanganato.com",
        "https://chapmanganato.com",
        "https://chapmanganato.to",
    ]
    headers = {"Referer": "https://readmanganato.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")
        title: str = soup.h1.next_element
        authors_genres = soup.select("td > a.a-h")
        authors: list[str] = []
        genres: list[str] = []
        for el in authors_genres:
            if el["href"].startswith("https://manganato.com/author"):
                authors.append(el.next_element)
            elif el["href"].startswith("https://manganato.com/genre"):
                genres.append(el.next_element)

        cover_img = next(iter(soup.select("span.info-image > img.img-loading")))
        cover_art: str = cover_img["src"]

        description = soup.select_one("#panel-story-info-description > h3").nextSibling.text.strip()

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(self._get_scripts(), "lxml")
        chapters = [BaseChapter(c.next_element, c["href"]) for c in soup.select("a.chapter-name")]
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        images = []
        for i in soup.find_all("img"):
            if not i["src"].startswith("https://readmanganato.com"):
                images.append(i["src"])
        return images

    @classmethod
    def url_to_id(cls, url: str) -> str:
        *_, last_item = filter(None, url.split("/"))
        return last_item

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(r"https://readmanganato.com/.*", url)
            or re.match(r"https://chapmanganato.com/.*", url)
            or re.match(r"https://chapmanganato.to/.*", url)
        )


def get_class() -> type[CommonSource]:
    return MangaNatoSource



#!/usr/bin/env python3

from pathlib import Path
from typing import cast

import typer

from . import (
    MD_METADATA_FILE,
    BaseChapter,
    BaseComic,
    BaseMetadata,
    ConvertFormats,
    ProcessConfig,
    ProcessOps,
    ProcessOptionMismatchError,
    SupportedProfiles,
    __version_str__,
    all_profiles,
    api,
    sources,
)
from .errors import ImageDownloadError

app = typer.Typer()


def cli_init_metadata_interactive() -> None:
    path: Path = typer.prompt("Folder path", default=Path.cwd(), type=Path).expanduser().resolve()

    try:
        comic = api.load(path)
    except FileNotFoundError:
        # expected if no metadata exists
        ...
    except IOError as err:
        typer.secho(f"{path} could not be found. Does it exist?", fg=typer.colors.RED)
        raise typer.Exit(1) from err
    else:
        typer.secho(
            f"There is already a comic at {path}. Remove md-metadata.json"
            "if you're sure you want to restart!",
            fg=typer.colors.RED,
        )
        raise typer.Exit(2)

    no_url_sentinel = "None"
    source_url: str = typer.prompt(
        "Automatically populate with the following source URL", default=no_url_sentinel
    )

    metadata = BaseMetadata("", [], "", [], "", "")
    chapters: list[BaseChapter] = []

    if source_url != no_url_sentinel:
        comic = cli_query(source_url)
        typer.echo(f"Found {comic.metadata.title} with {len(comic.chapters)} chapters")
        metadata = comic.metadata
        chapters = comic.chapters

    metadata.title = typer.prompt("Title", default=metadata.title or None).strip()
    metadata.__post_init__()  # regenerate slug

    metadata.authors = [
        s.strip()
        for s in typer.prompt(
            "Author(s) (comma-separated)",
            default=", ".join(metadata.authors) or None,
        ).split(",")
    ]
    metadata.genres = [
        s.strip()
        for s in typer.prompt(
            "Genre(s) (comma-separated)",
            default=", ".join(metadata.genres) or None,
        ).split(",")
    ]
    metadata.cover_art = typer.prompt(
        "Cover art URL (enter 'EXISTS' if cover.png/jpg already exists)",
        default=metadata.cover_art or None,
    ).strip()
    metadata.description = typer.prompt("Description", default=metadata.description or None).strip()
    metadata.url = metadata.url

    typer.secho("Metadata collected, now adding chapters...", fg=typer.colors.GREEN)

    folders_in_cd = sorted(f for f in path.iterdir() if f.is_dir())
    delta = len(folders_in_cd)  # #folders - #chapters
    if chapters:
        # try to match up chapters with existing files
        for folder, chapter in zip(folders_in_cd, chapters, strict=False):
            chapter.slug = folder.stem

        # print out the matches
        typer.secho("Matches found:", fg=typer.colors.GREEN)
        for folder, chapter in zip(folders_in_cd, chapters, strict=False):
            typer.echo(f"  {folder.stem} -> {chapter.title}")

        delta = len(folders_in_cd) - len(chapters)

    if delta > 0:
        res = typer.prompt(
            f"Matching stalled with {delta} extra local chapters. Attempt to resolve? [Y/n]",
            default=True,
            show_choices=True,
            show_default=False,
        )
        if res:
            # if there are more folders than chapters, add them as new chapters
            for folder in folders_in_cd[len(chapters) :]:
                chapters.append(BaseChapter(folder.stem, "", folder.stem))

            # print out new matches
            typer.secho("New chapters:", fg=typer.colors.GREEN)
            for folder, _ in zip(folders_in_cd, chapters, strict=False):
                typer.echo(f"  NEW: {folder.stem}")

    res = typer.prompt(
        "Matching completed. Finish and save? [Y/n]",
        default=True,
        show_choices=True,
        show_default=False,
    )

    if not res:
        raise typer.Abort()

    typer.secho(
        f"All done! Saving metadata to {path / MD_METADATA_FILE}...",
        fg=typer.colors.GREEN,
    )
    api.init_parse_comic(path, BaseComic(metadata, chapters), metadata.cover_art != "EXISTS")


def cli_query(url: str) -> BaseComic:
    typer.echo(f"Searching sources for {url}")

    try:
        comic = api.query(url)
    except ValueError as err:
        typer.secho("Could not match URL with available sources.", fg=typer.colors.RED)
        raise typer.Exit(1) from err
    return comic


def cli_convert(
    comic_path: Path,
    target_format: ConvertFormats,
    dest_folder: Path = Path.cwd(),
    remove_after: bool = False,
    split_by_chapters: bool = False,
) -> None:
    comic = api.load(comic_path)
    iterator = api.convert_progress(
        comic_path, target_format, dest_folder, remove_after, split_by_chapters
    )

    is_single_conversion = comic_path.is_dir()

    try:
        len_first_conv = cast(int, next(iterator))
    except RuntimeError as err:
        # handle kindlegen errors
        typer.secho(str(err), fg=typer.colors.RED)
        raise typer.Abort() from None

    len_second_conv = -1

    first_convert_message = (
        f"Packing {target_format.value}(s)" if is_single_conversion else "Pre-converting comic"
    )

    try:
        with typer.progressbar(
            iterator,
            length=len_first_conv,
            label=first_convert_message,
        ) as progress:
            for res in progress:
                if isinstance(res, int):
                    len_second_conv = res
                    break
                if split_by_chapters:
                    progress.label = res
    except RuntimeError as err:
        # handle kindlegen errors
        typer.secho(str(err), fg=typer.colors.RED)
        raise typer.Abort() from None

    if not is_single_conversion:
        # it *should* be guaranteed len_second_conv exists
        try:
            with typer.progressbar(
                iterator,
                length=len_second_conv,
                label=f"Packing {target_format.value}",
            ) as progress:
                for _ in progress:
                    ...
        except RuntimeError as err:
            # handle kindlegen errors
            typer.secho(str(err), fg=typer.colors.RED)
            raise typer.Abort() from None

    if not split_by_chapters:
        dest_file = dest_folder / f"{comic.metadata.title_slug}.{target_format.value}"
        typer.secho(f"Successfully converted to {dest_file}", fg=typer.colors.GREEN)
    else:
        typer.secho(f"Successfully converted to {dest_folder}", fg=typer.colors.GREEN)


def cli_process(comic_path: Path, options: list[ProcessOps], config: ProcessConfig) -> None:
    if ProcessOps.NO_POSTPROCESSING in options:
        return

    try:
        comic = api.load(comic_path)
    except FileNotFoundError as err:
        typer.secho(
            f"Comic not found at {comic_path}, is md-metadata.json missing?",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1) from err

    typer.secho(f"Applying processing options: {', '.join(options)}", fg=typer.colors.GREEN)
    try:
        with typer.progressbar(
            api.process_progress(comic_path, options, config),
            length=len(comic.chapters),
            label="Processing",
        ) as progress:
            for _ in progress:
                pass
    except ProcessOptionMismatchError as err:
        typer.secho(f"Could not apply processing options: {err}", fg=typer.colors.RED)
        raise typer.Exit(1) from err


@app.command(no_args_is_help=True)
def convert(
    convert_to: ConvertFormats,
    folder_path: Path,
    dest: Path = typer.Option(
        Path.cwd(),
        "--dest",
        "-d",
        help="The folder to save the converted file to.",
    ),
    remove_after: bool = typer.Option(
        False,
        "--remove-after",
        "-r",
        help="Remove the the original folder after conversion",
    ),
    split_by_chapters: bool = typer.Option(
        False,
        "--split-by-chapters",
        "-b",
        help="Instead of returning one large comic file, create one comic"
        "file for each chapter (applies only to Mandown-created comic folders)",
    ),
) -> None:
    """
    Convert a comic OR comic folder into CBZ/EPUB/PDF.

    eg. To convert to CBZ:
    mandown convert cbz /path/to/comic/folder

    eg. To convert an existing PDF comic to EPUB:
    mandown convert epub /path/to/comic.pdf
    """
    typer.echo(f"Converting to {convert_to}...")
    cli_convert(folder_path, convert_to, dest, remove_after, split_by_chapters)


@app.command(no_args_is_help=True)
def process(
    options: list[ProcessOps],
    folder_path: Path = typer.Argument(Path.cwd()),
    target_size: tuple[int, int] | None = typer.Option(
        (0, 0),
        "--target-size",
        "-z",
        min=0,
        show_default=False,
        help="RESIZE ONLY: The target size (width, height) (cannot be used with `profile`)",
    ),
    size_profile: str | None = typer.Option(
        None,
        "--profile",
        "-o",
        help="RESIZE ONLY: The device profile to use (cannot be used with `target-size`)",
    ),
) -> None:
    """
    Process a comic folder in-place.

    eg. To trim borders and resize to 800x1200:
    mandown process trim_borders resize -z 800 1200

    eg. To split double pages and resize to a Kindle Paperwhite 2 profile:
    mandown process split_double_pages resize -o paper

    All profiles can be listed with "mandown --list-profiles".
    """
    # work around typer bug (see mandown get)
    if target_size == (0, 0):
        target_size = None

    size_profile = cast(SupportedProfiles | None, size_profile)

    try:
        config = ProcessConfig(
            target_size=target_size,
            output_profile=size_profile,
        )
    except Exception as err:
        typer.secho(f"Could not apply processing options: {err}", fg=typer.colors.RED)
        raise typer.Exit(1) from err
    cli_process(folder_path, options, config)


@app.command(no_args_is_help=True)
def get(
    url: str,
    dest: Path = typer.Argument(Path.cwd(), help="The destination folder to download to."),
    convert_to: ConvertFormats = typer.Option(
        "none", "--convert", "-c", help="The format to download the comic as"
    ),
    start: int | None = typer.Option(
        None,
        "--start",
        "-s",
        help="The first chapter to download [default: first found]",
    ),
    end: int | None = typer.Option(
        None, "--end", "-e", help="The last chapter to download [default: last found]"
    ),
    maxthreads: int = typer.Option(
        4,
        "--threads",
        "-t",
        help="The maximum number of images to download in parallel",
    ),
    processing_options: list[ProcessOps] | None = typer.Option(
        [],
        "--process",
        "-p",
        help="Image processing options (in-place)",
        case_sensitive=True,
    ),
    target_size: tuple[int, int] | None = typer.Option(
        (0, 0),
        "--target-size",
        "-z",
        show_default=False,
        min=0,
        help="IF PROCESSING AND RESIZING: The target size (width, height)",
    ),
    size_profile: str | None = typer.Option(
        None,
        "--profile",
        "-o",
        help="IF PROCESSING AND RESIZING: The device profile to use",
    ),
    remove_after: bool = typer.Option(
        False,
        "--remove-after",
        "-r",
        help="IF CONVERTING: Remove the downloaded folder after converting",
    ),
    split_by_chapters: bool = typer.Option(
        False,
        "--split-by-chapters",
        "-b",
        help="IF CONVERTING: Instead of returning one large comic file, create one comic"
        "file for each chapter (applies only to Mandown-created comic folders)",
    ),
) -> None:
    """
    Download from a URL chapters start_chapter to end_chapter.
    Defaults to the first chapter and last chapter, respectively
    in the working directory.

    eg. To download all chapters of a comic to the current directory:
    mandown get https://website.com/comic/1234

    eg. To download chapters 1-10 of a comic to the current directory:
    mandown get https://website.com/comic/1234 -s 1 -e 10

    eg. To convert all chapters of a comic to EPUB:
    mandown get https://website.com/comic/1234 -c epub

    eg. To convert all chapters of a comic to CBZ after trimming borders:
    mandown get https://website.com/comic/1234 -c cbz -p trim_borders
    """
    # work around typer bug (optional of tuples is not parsed correctly)
    if target_size == (0, 0):
        target_size = None

    if not dest.is_dir():
        raise ValueError(f"{dest} is not a valid folder path.")

    size_profile = cast(SupportedProfiles | None, size_profile)

    # get and save metadata
    comic = cli_query(url)

    typer.echo(
        f'Found comic "{comic.metadata.title}" from source {sources.get_class_for(url).name}',
    )

    # get processing range
    start_chapter = start or 1
    end_chapter = end or len(comic.chapters)

    # zero-index
    start_chapter -= 1

    comic.set_chapter_range(start=start_chapter, end=end_chapter)

    # download
    typer.echo(f"Downloading {end_chapter - start_chapter} chapter(s)...")
    try:
        with typer.progressbar(
            api.download_progress(comic, dest, threads=maxthreads),
            length=len(comic.chapters),
        ) as progress:
            for title in progress:
                progress.label = title
    except ImageDownloadError as err:
        typer.secho(
            "Some image links on the host site were broken, ignoring...",
            fg=typer.colors.BRIGHT_YELLOW,
        )
        typer.secho(f"Error: {err}", fg=typer.colors.RED)

    full_dest_folder = dest.absolute() / comic.metadata.title_slug
    typer.secho(
        f"Successfully downloaded {end_chapter - start_chapter} chapter(s) to {full_dest_folder}.",
        fg=typer.colors.GREEN,
    )

    # process
    if processing_options:
        try:
            config = ProcessConfig(
                target_size=target_size,
                output_profile=size_profile,
            )
        except Exception as err:
            typer.secho(f"Could not apply processing options: {err}", fg=typer.colors.RED)
            raise typer.Exit(1) from err

        cli_process(dest / comic.metadata.title_slug, processing_options, config)

    # convert
    if convert_to != ConvertFormats.NONE:
        cli_convert(
            dest / comic.metadata.title_slug,
            convert_to,
            dest,
            remove_after,
            split_by_chapters,
        )


@app.command(name="init-metadata")
def init_metadata(
    path: Path | None = typer.Argument(None, help="The folder to initialise"),
    source_url: str | None = typer.Argument(None, help="The url to get metadata from"),
    download_cover: bool = typer.Option(
        False,
        "--download-cover",
        "-d",
        help="If --source-url is passed: Download the cover image from the source.",
    ),
) -> None:
    """
    Initialise a folder with metadata to be converted with Mandown, optionally
    fetching metadata from an internet source. Pass with no arguments to
    start an interactive session.

    eg. mandown init-metadata /path/to/folder https://website.com/comic/1234 --download-cover
    """

    if path is None:
        # interactive session
        return cli_init_metadata_interactive()

    if (path / MD_METADATA_FILE).is_file():
        return typer.echo("Metadata already found. Please remove it to create new metadata.")

    try:
        if source_url is not None:
            donor = api.query(source_url)
            comic = api.init_parse_comic(path, donor, download_cover)
        else:
            comic = api.init_parse_comic(path)
    except AttributeError as err:
        typer.secho("--download-cover must be used with --source-url.")
        raise typer.Exit(1) from err
    except ValueError as err:
        typer.secho("Source not found.")
        raise typer.Exit(1) from err

    typer.secho(f"Found {comic.metadata.title}:", fg=typer.colors.BRIGHT_GREEN)
    typer.echo(comic)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def callback(
    version: bool | None = typer.Option(
        None,
        "--version",
        "-v",
        is_eager=True,
        help="Display the current version of mandown",
    ),
    supported_sites: bool | None = typer.Option(
        None,
        "--supported-sites",
        is_eager=True,
        help="Output a list of domains supported by mandown",
    ),
    list_profiles: bool = typer.Option(
        False,
        "--list-profiles",
        "-l",
        help="List available device profiles and details",
        is_eager=True,
    ),
) -> None:
    if version:
        typer.echo(f"mandown {__version_str__}")
        raise typer.Exit()

    if supported_sites:
        for source in sources.get_all_classes():
            typer.echo(f"{source.name}: {', '.join(source.domains)}")
        raise typer.Exit()

    if list_profiles:
        typer.echo("Available profiles:")
        typer.echo(
            "\n".join(f" - {profile.name}: {profile.id!r}" for profile in all_profiles.values())
        )
        raise typer.Exit()


def main() -> None:
    app()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 5.15.4
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *


class Ui_Widget(object):
    def setupUi(self, Widget):
        if not Widget.objectName():
            Widget.setObjectName("Widget")
        Widget.resize(800, 600)
        self.verticalLayout_2 = QVBoxLayout(Widget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_4 = QLabel(Widget)
        self.label_4.setObjectName("label_4")

        self.horizontalLayout_2.addWidget(self.label_4)

        self.text_source = QLineEdit(Widget)
        self.text_source.setObjectName("text_source")

        self.horizontalLayout_2.addWidget(self.text_source)

        self.button_from_url = QPushButton(Widget)
        self.button_from_url.setObjectName("button_from_url")
        icon = QIcon(QIcon.fromTheme("search"))
        self.button_from_url.setIcon(icon)

        self.horizontalLayout_2.addWidget(self.button_from_url)

        self.button_from_folder = QPushButton(Widget)
        self.button_from_folder.setObjectName("button_from_folder")
        icon1 = QIcon(QIcon.fromTheme("folder"))
        self.button_from_folder.setIcon(icon1)

        self.horizontalLayout_2.addWidget(self.button_from_folder)

        self.verticalLayout_2.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_5 = QLabel(Widget)
        self.label_5.setObjectName("label_5")

        self.horizontalLayout_4.addWidget(self.label_5)

        self.text_dest = QLineEdit(Widget)
        self.text_dest.setObjectName("text_dest")

        self.horizontalLayout_4.addWidget(self.text_dest)

        self.pushButton_3 = QPushButton(Widget)
        self.pushButton_3.setObjectName("pushButton_3")
        self.pushButton_3.setIcon(icon1)

        self.horizontalLayout_4.addWidget(self.pushButton_3)

        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_2 = QLabel(Widget)
        self.label_2.setObjectName("label_2")

        self.verticalLayout.addWidget(self.label_2)

        self.label_image = QLabel(Widget)
        self.label_image.setObjectName("label_image")

        self.verticalLayout.addWidget(self.label_image)

        self.label_metadata = QLabel(Widget)
        self.label_metadata.setObjectName("label_metadata")

        self.verticalLayout.addWidget(self.label_metadata)

        self.horizontalLayout.addLayout(self.verticalLayout)

        self.verticalLayout_4 = QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label = QLabel(Widget)
        self.label.setObjectName("label")

        self.verticalLayout_4.addWidget(self.label)

        self.chapter_table = QTableWidget(Widget)
        self.chapter_table.setObjectName("chapter_table")

        self.verticalLayout_4.addWidget(self.chapter_table)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_3 = QLabel(Widget)
        self.label_3.setObjectName("label_3")

        self.horizontalLayout_3.addWidget(self.label_3)

        self.radio_no_convert = QRadioButton(Widget)
        self.radio_no_convert.setObjectName("radio_no_convert")

        self.horizontalLayout_3.addWidget(self.radio_no_convert)

        self.radio_convert_epub = QRadioButton(Widget)
        self.radio_convert_epub.setObjectName("radio_convert_epub")

        self.horizontalLayout_3.addWidget(self.radio_convert_epub)

        self.radio_convert_pdf = QRadioButton(Widget)
        self.radio_convert_pdf.setObjectName("radio_convert_pdf")

        self.horizontalLayout_3.addWidget(self.radio_convert_pdf)

        self.radio_convert_cbz = QRadioButton(Widget)
        self.radio_convert_cbz.setObjectName("radio_convert_cbz")

        self.horizontalLayout_3.addWidget(self.radio_convert_cbz)

        self.verticalLayout_4.addLayout(self.horizontalLayout_3)

        self.horizontalLayout.addLayout(self.verticalLayout_4)

        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_progress = QLabel(Widget)
        self.label_progress.setObjectName("label_progress")

        self.horizontalLayout_5.addWidget(self.label_progress)

        self.progress_bar = QProgressBar(Widget)
        self.progress_bar.setObjectName("progress_bar")
        self.progress_bar.setValue(24)

        self.horizontalLayout_5.addWidget(self.progress_bar)

        self.verticalLayout_2.addLayout(self.horizontalLayout_5)

        self.button_start = QPushButton(Widget)
        self.button_start.setObjectName("button_start")

        self.verticalLayout_2.addWidget(self.button_start)

        self.retranslateUi(Widget)

        QMetaObject.connectSlotsByName(Widget)

    # setupUi

    def retranslateUi(self, Widget):
        Widget.setWindowTitle(QCoreApplication.translate("Widget", "Widget", None))
        self.label_4.setText(QCoreApplication.translate("Widget", "Source Folder / URL:", None))
        self.button_from_url.setText(QCoreApplication.translate("Widget", "Search URL", None))
        self.button_from_folder.setText(QCoreApplication.translate("Widget", "Open Folder", None))
        self.label_5.setText(QCoreApplication.translate("Widget", "Save to:", None))
        self.pushButton_3.setText(QCoreApplication.translate("Widget", "Browse", None))
        self.label_2.setText(QCoreApplication.translate("Widget", "Metadata", None))
        self.label.setText(QCoreApplication.translate("Widget", "Chapters", None))
        self.label_3.setText(QCoreApplication.translate("Widget", "Convert?", None))
        self.radio_no_convert.setText(QCoreApplication.translate("Widget", "No", None))
        self.radio_convert_epub.setText(QCoreApplication.translate("Widget", "EPUB", None))
        self.radio_convert_pdf.setText(QCoreApplication.translate("Widget", "PDF", None))
        self.radio_convert_cbz.setText(QCoreApplication.translate("Widget", "CBZ", None))
        self.label_progress.setText(QCoreApplication.translate("Widget", "Downloading...", None))
        self.button_start.setText(QCoreApplication.translate("Widget", "Start!", None))

    # retranslateUi



import requests as RealRequests

USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36"  # noqa: E501


class requests:
    @staticmethod
    def get(url: str) -> RealRequests.Response:
        return RealRequests.get(url, headers={"User-Agent": USER_AGENT}, timeout=5)



"""
Source file for manhuaes.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class ManhuaESSource(CommonSource):
    name = "ManhuaAZ"
    domains = ["https://manhuaaz.com"]
    # headers = {"Referer": "https://mangakakalot.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self.url = f"https://manhuaaz.com/manga/{self.id}"
        self.numerical_id: str | None = None
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")
        self.numerical_id = soup.select_one("input.rating-post-id")["value"]

        title: str = soup.select_one(".manga-title-badges + h1").text.strip()
        authors: list[str] = [
            string.strip()
            for string in soup.select_one(".author-content").text.split(", ")
            if string.strip() != "Updating"
        ]
        genres: list[str] = [
            string.strip()
            for string in soup.select_one(".genres-content").text.split(", ")
            if string.strip() != "Updating"
        ]
        cover_art: str = soup.select_one(".summary_image img")["data-src"]
        description: str = soup.select_one(".summary__content").text.strip()

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        if not self.numerical_id:
            soup = BeautifulSoup(self._get_scripts(), "lxml")
            self.numerical_id = soup.select_one("input.rating-post-id")["value"]

        soup = BeautifulSoup(
            requests.post(
                "https://manhuaaz.com/wp-admin/admin-ajax.php",
                data={"action": "manga_get_chapters", "manga": self.numerical_id},
            ).text,
            "lxml",
        )
        chapters = [
            BaseChapter(c.text.strip(), c["href"]) for c in soup.select(".wp-manga-chapter a")
        ]
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        return [el["data-src"] for el in soup.select("img.wp-manga-chapter-img")]

    @classmethod
    def url_to_id(cls, url: str) -> str:
        items = list(filter(None, url.split("/")))
        # e.g., for https://manhuaes.com/manga/the-undefeatable-swordsman-all/chapter-166/"
        # we want 'the-undefeatable-swordsman-all'
        return items[3]

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://manhua(es|az).com/manga/.*", url))


def get_class() -> type[CommonSource]:
    return ManhuaESSource



from ..base import BaseChapter, BaseMetadata


class BaseSource:
    """
    An object representing network state for fetching data of a comic.
    """

    name = "Source name goes here"
    domains = ["Source domains goes here"]
    headers: dict[str, str] = {}

    _metadata: BaseMetadata | None = None
    _chapters: list[BaseChapter] = []

    def __init__(self, url: str):
        self.url = url

    @property
    def metadata(self) -> BaseMetadata:
        """
        Return the metadata of the comic, fetching and caching it if necessary.
        """
        if self._metadata:
            return self._metadata
        self._metadata = self.fetch_metadata()
        return self._metadata

    @property
    def chapters(self) -> list[BaseChapter]:
        """
        Return the chapter list of the comic, fetching and caching it if necessary.
        """
        if self._chapters:
            return self._chapters
        self._chapters = self.fetch_chapter_list()
        return self._chapters

    def fetch_metadata(self) -> BaseMetadata:
        """
        Fetch and return title, author, and url of the comic.
        """
        raise NotImplementedError("Metadata getter not overridden")

    def fetch_chapter_list(self) -> list[BaseChapter]:
        """
        Fetch and return a list of chapter titles and their URL in ascending order.
        """
        raise NotImplementedError("Chapter list getter not overridden")

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        """
        Given a chapter link, fetch and return a list of image URLs in ascending
        order for that chapter.
        """
        raise NotImplementedError("Image URL getter not overridden")

    @staticmethod
    def check_url(url: str) -> bool:
        """
        Returns whether the url given matches that of the site and is processable
        """
        raise NotImplementedError("URL checker not overridden")


def get_class() -> type[BaseSource]:
    return BaseSource



# pylint: disable=no-member,invalid-name,pointless-string-statement
"""
Generate mainwin.py while in mandown/ui/ with
`uic -g python -o mainwin.py form.ui`
"""

import sys
from pathlib import Path

import requests
from PySide6.QtGui import QImage, QPixmap

import mandown

from .. import __version_str__
from ..comic import BaseComic
from .mainwin import Ui_Widget

from PySide6.QtWidgets import (  # isort: skip
    QApplication,
    QFileDialog,
    QMessageBox,
    QTableWidgetItem,
    QWidget,
)


COVER_IMG_HEIGHT = 400


class MandownQtUi(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.ui = Ui_Widget()
        self.ui.setupUi(self)
        self.init_hooks()

        self.ui.progress_bar.setDisabled(True)
        self.ui.progress_bar.setValue(0)
        self.ui.label_progress.setText("Waiting")

        self.ui.label_metadata.setWordWrap(True)

        self.setWindowTitle(f"Mandown {__version_str__}")

        self.source_path: str | None = None
        self.source_url: str | None = None
        self.dest_path: str | None = None
        self.chapters = None
        self._comic: BaseComic | None = None
        self.img_cover: QImage | None = None

        self.show()

    def init_hooks(self) -> None:
        # select "No" for default convert
        self.ui.radio_no_convert.toggle()
        # button: Open Folder
        self.ui.button_from_folder.clicked.connect(self.hook_from_folder)
        # button: Search URL
        self.ui.button_from_url.clicked.connect(self.hook_from_url)
        # button: Browse
        self.ui.pushButton_3.clicked.connect(self.hook_set_dest)
        # button: Start!
        self.ui.button_start.clicked.connect(self.hook_go)

    def metadata_to_table(self) -> list[str]:
        metadata = self.comic.metadata
        return [
            f"<b>Title:</b> {metadata.title}",
            f"<b>Author(s):</b> {', '.join(metadata.authors)}",
            f"<b>Genre(s):</b> {', '.join(metadata.genres)}",
            "",
            metadata.description,
        ]

    def chapters_to_table(self) -> list:
        # TODO: make them VIEWS / LAYOUTS not widgets
        # because i can't add checkboxes
        return [c.title for c in self.comic.chapters]

    @property
    def comic(self) -> BaseComic | None:
        return self._comic

    @comic.setter
    def comic(self, comic: BaseComic) -> None:
        self._comic = comic

        self.ui.label_metadata.setText("<br />".join(self.metadata_to_table()))

        for i, chap in enumerate(self.comic.chapters):
            self.ui.chapter_table.setItem(i, 0, QTableWidgetItem(""))
            self.ui.chapter_table.setItem(i, 1, QTableWidgetItem(chap.title))

        # refresh screen w/metadata et al here
        if comic.metadata.cover_art:
            r = requests.get(comic.metadata.cover_art)
            self.img_cover = QPixmap(QImage.fromData(r.content).scaledToHeight(COVER_IMG_HEIGHT))
            self.ui.label_image.setPixmap(self.img_cover)

    @property
    def text_dest_full(self) -> str:
        return str(Path(self.ui.text_dest.text()) / self.comic.metadata.title)

    """
    Hooks go here!
    """

    def hook_from_folder(self) -> None:
        # TODO: directly look for md-metadata.json instead
        self.source_path = QFileDialog.getExistingDirectory(self, "Open Comic Folder", "~")
        self.ui.text_source.setText(self.source_path)
        self.ui.text_dest.setText(str(Path(self.source_path).parent))

        # load metadata
        try:
            self.comic = mandown.load(self.source_path)
        except FileNotFoundError:
            res = QMessageBox.critical(
                self,
                "Invalid Folder",
                "Folder is not a Mandown-created comic folder: Could not find md-metadata.json.",
                QMessageBox.Ok,
            )

    def hook_from_url(self) -> None:
        active_url = self.ui.text_source.text()
        try:
            comic = mandown.query(active_url)
        except ValueError:
            res = QMessageBox.critical(
                self,
                "Unknown Comic",
                "Could not find comic: URL did not match any sources.",
                QMessageBox.Ok,
            )
        self.comic = comic

    def hook_set_dest(self) -> None:
        self.dest_path = QFileDialog.getExistingDirectory(self, "Set Destination Folder", "~")
        self.ui.text_dest.setText(self.dest_path)

    def hook_go(self) -> None:
        if not self.ui.text_source.text():
            # if empty
            res = QMessageBox.critical(
                self,
                "Missing source",
                "Please select a source folder or URL",
                QMessageBox.Ok,
            )

        if not self.ui.text_dest.text():
            # if empty
            res = QMessageBox.critical(
                self,
                "Missing destination",
                "Please select a destination directory.",
                QMessageBox.Ok,
            )
            return

        if not self.comic:
            res = QMessageBox.critical(
                self,
                "Something went wrong",
                "ERROR: if not self.comic check failed. You should never see this message.",
            )
            return

        # heavy bits
        self.ui.label_progress.setText("Downloading")
        max_size = len(self.comic.chapters)
        self.ui.progress_bar.setDisabled(False)
        for i, _ in enumerate(
            mandown.download_progress(
                self.comic,
                self.ui.text_dest.text(),
                threads=4,
            ),
            start=1,
        ):
            # TODO: this hangs the program so move it to a QThread
            self.ui.progress_bar.setValue(int(i / max_size * 100))

        # convert
        if not self.ui.radio_no_convert.isChecked():
            target = "none"
            if self.ui.radio_convert_cbz.isChecked():
                target = "cbz"
            elif self.ui.radio_convert_epub.isChecked():
                target = "epub"
            elif self.ui.radio_convert_pdf.isChecked():
                target = "pdf"

            self.ui.label_progress.setText("Converting")
            self.ui.progress_bar.setValue(1)
            for i, text in enumerate(
                mandown.convert_progress(
                    self.comic,
                    self.text_dest_full,
                    target,
                    self.text_dest_full,
                ),
                start=1,
            ):
                self.ui.label_progress.setText(text)
                self.ui.progress_bar.setValue(int(i / max_size * 100))

        res = QMessageBox.information(self, "Done", "All operations complete.", QMessageBox.Ok)

        self.ui.progress_bar.setDisabled(True)


def main() -> None:
    app = QApplication([])
    wid = MandownQtUi()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from .profiles import SupportedProfiles, all_profiles

try:
    from PIL import Image, ImageChops

    if not hasattr(Image, "Resampling"):  # Pillow<9.0
        Image.Resampling = Image
except ImportError:
    if not TYPE_CHECKING:

        class Image:
            class Image:
                pass

        class ImageChops:
            pass


@dataclass(kw_only=True, slots=True)
class ProcessConfig:
    """
    A class for storing configuration for processing images.

    :param `target_size`: The target size for the image. Only
    used if `resize` is enabled. Mutually exclusive with `output_profile`.
    :param `output_profile`: The output size profile to use for
    the image. Only used if `resize` is enabled. Mutually exclusive with `target_size`.
    :raises `ValueError`: If incompatible options are supplied
    :raises `KeyError`: `output_profile` is not a real key (see mandown/processor/profiles.py).
    """

    target_size: tuple[int, int] | None = None
    """
    The target size for the image. Only used if `resize`
    is enabled. Mutually exclusive with `output_profile`.
    """

    output_profile: SupportedProfiles | None = None
    """
    The output size profile to use for the image. Only
    used if `resize` is enabled. Mutually exclusive with `target_size`.
    """

    def __post_init__(self) -> None:
        if self.target_size is not None and self.output_profile is not None:
            raise ValueError("Only one of `target_size` or `output_profile` can be specified.")

        if self.output_profile is not None:
            if self.output_profile not in all_profiles:
                raise KeyError(
                    f"Invalid output profile: {self.output_profile}. See mandown."
                    "all_profiles or mandown --list-profiles for a list of valid profiles."
                )
            self.target_size = all_profiles[self.output_profile].size


class ProcessContainer:
    """
    Add processing functions here!
    Name them by their Enum string in ProcessOps.
    They should all return a Image.Image | tuple[Image.Image] | None
    depending on if they can return multiple images.

    The first image returned will replace the current image,
    all other ones will be suffixed with `a`. If `None` is returned,
    the image will not be altered.
    """

    def __init__(self, config: ProcessConfig | None = None):
        self.config = config or ProcessConfig()

    def rotate_double_pages(
        self,
        image: Image.Image,
    ) -> Image.Image | None:
        """
        Rotate the image 90 degrees if it is a double page so it fits on the screen.
        """
        width, height = image.size
        if width > height:
            return image.rotate(90, expand=1)
        return None

    def split_double_pages(self, image: Image.Image) -> tuple[Image.Image, Image.Image] | None:
        """
        Split the image into two separate images if it is a double page.
        """
        width, height = image.size
        if not width > height:
            return None

        left = image.crop((0, 0, int(width / 2), height))
        right = image.crop((int(width / 2), 0, width, height))
        return (left, right)

    def trim_borders(self, image: Image.Image) -> Image.Image | None:
        """
        Trim the borders of the image.
        """
        bg = Image.new(image.mode, image.size, image.getpixel((0, 0)))
        diff = ImageChops.difference(image, bg)
        diff = ImageChops.add(diff, diff, 2.0, -100)
        bbox = diff.getbbox()
        if bbox:
            return image.crop(bbox)
        return None

    def resize(
        self,
        image: Image.Image,
    ) -> Image.Image | None:
        """
        Resize the image to a maximum width and height.
        """
        target_size = self.config.target_size

        if target_size is None or image.size == target_size:
            return None
        return image.resize(target_size, resample=Image.Resampling.LANCZOS)


class OutputProcessContainer:
    """
    Add output processing functions here!
    Name them by their Enum string in ProcessOps.
    They should all return None.
    """

    def __init__(self, image: Image.Image, filename: str | Path) -> None:
        self.image = image
        self.filename = filename

    def default(self) -> None:
        """
        Write the image to disk.
        """
        self.image.save(self.filename)

    def webp_to_png(self) -> None:
        """
        Convert any WEBP images to PNG.
        """
        if Path(self.filename).suffix != ".webp":
            return

        path = Path(self.filename)
        self.image.save(path.with_suffix(".png"), "PNG")
        path.unlink()



"""
Source file for bato.to
"""

import html
import json
import re
from typing import cast

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class BatotoSource(CommonSource):
    name = "Bato.to"
    domains = ["https://bato.to"]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)  # we make a GET request here which is not ideal
        self.url = f"https://bato.to/title/{self.id}"
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        cover_art_el = soup.select_one("img.not-prose")
        title = cast(str, cover_art_el["alt"])
        cover_art = cast(str, cover_art_el["src"])
        authors = list({strip_parentheses(item.text) for item in soup.select("div.mt-2 a")})
        genres = [
            strip_parentheses(g.text)
            for g in soup.select(
                "div.space-y-2 > div.flex.items-center.flex-wrap > span > :nth-child(1)"
            )
        ]
        description = soup.select_one("astro-island > div > .prose > .limit-html-p").text.strip()

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        chapters: list[BaseChapter] = []
        for item in soup.select('div[name="chapter-list"] div.space-x-1 > a:nth-child(1)'):
            link = cast(str, item["href"])
            title = item.text
            chapters.append(BaseChapter(title, f"https://bato.to{link}?load=2"))
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")

        """
        It looks something like:
        {'pageOpts': [0, {'load': [0, '2'], 'marg': [0, '0'], 'zoom': [0, '0']}], 'imageFiles': [1,
        '[[0,"https://xfs-n03.xfsbb.com/comic/7006/fbf/65ac7d07a811b44f5e9abfbf/45869815_2560_2824_374174.webp"]]'
        ], 'urlP': [0, 0]}
        """
        data = json.loads(
            html.unescape(
                soup.select_one('astro-island[component-url^="/_astro/ImageList"]')["props"]
            )
        )

        images: list[str] = []
        for image in json.loads(data["imageFiles"][1]):
            images.append(image[1])
        return images

    @classmethod
    def url_to_id(cls, url: str) -> str:
        items = list(filter(None, url.split("/")))

        # e.g., for https://bato.to/title/115663-my-not-so-fair-lady-is-doomed-but-not-if-i-can-help-it-official/2950514-ch_15
        # we want "115663-my-not-so-fair-lady-is-doomed-but-not-if-i-can-help-it-official"

        return items[3]

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://bato.to/title/.*", url))


def get_class() -> type[CommonSource]:
    return BatotoSource


def strip_parentheses(text: str) -> str:
    index = text.find("(")
    if index != -1:
        return text[:index].strip()
    return text.strip()



"""
Source file for kuaikanmanhua.com
"""
# pylint: disable=invalid-name

import json
import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class KuaiKanManhuaSource(CommonSource):
    name = "快看漫画 (Kuaikan Manhua)"
    domains = ["https://www.kuaikanmanhua.com"]
    # headers = {"Referer": "https://www.kuaikanmanhua.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)  # we make a GET request here which is not ideal
        self.url = f"https://www.kuaikanmanhua.com/web/topic/{self.id}"
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        title: str = soup.select_one("h3.title").text
        authors: list[str] = [soup.select_one("div.nickname").text.strip()]
        genres: list[str] = []  # no genres accessible from the list page on this site
        cover_art: str = soup.select_one("img.img")["src"]

        description = soup.select_one("div.detailsBox > p").text.strip().lstrip("漫画简介:").strip()

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        text = self._get_scripts()
        soup = BeautifulSoup(text, "lxml")
        array_start = text.index("Array(")
        next_https = text.index('"https:', array_start)
        previous_quote = text.rindex('"', array_start, next_https)
        prev_prev_quote = text.rindex(
            '"', array_start, previous_quote
        )  # this should be the quote of the first el in what we want

        end = text.index("{}", next_https)
        last_comma = text.rindex(",", next_https, end)

        constructed = f"[{text[prev_prev_quote:last_comma]}]"
        strings: list[str | int] = json.loads(constructed)

        # Array(7),"都市","热血","投稿","1.起源","https://...
        # we want to get "1.起源","https://...
        # but "都市","热血","投稿" may not be there

        # the first chapter does not have its id in the array
        first_chapter_id = int(soup.select_one("a.firstBtn")["data-href"].split("/")[-1])
        strings.insert(0, first_chapter_id)

        # if it is an int, it must be the chapter id, then the title
        chapters: list[BaseChapter] = []
        for i, string in enumerate(strings):
            if isinstance(string, int) and string >= 459:  # magic number from binary search
                # there are some weird numbers like 257 that pop up and are not chapters
                chapters.append(
                    BaseChapter(
                        strings[i + 1], f"https://www.kuaikanmanhua.com/webs/comic-next/{string}"
                    )
                )
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        text = requests.get(chapter.url).text
        js_start = text.index("config:{_app:")
        next_sign = text.index("?sign=", js_start)
        prev_quote = text.rindex('"', js_start, next_sign)
        last_sign = text.rindex("?sign=", js_start)
        last_comma = text.index(",", last_sign)
        constructed = f"[{text[prev_quote:last_comma]}]"
        strings: list[str] = json.loads(constructed)
        return [s for s in strings if isinstance(s, str)]  # there may be ints

    @classmethod
    def url_to_id(cls, url: str) -> str:
        items = list(filter(None, url.split("/")))
        # e.g., for https://www.kuaikanmanhua.com/web/topic/16222/
        # we want '16222'
        # for https://www.kuaikanmanhua.com/webs/comic-next/540156
        # we want to crawl the page and find the link back to the ID

        if "topic" in items:
            return int(items[4])
        soup = BeautifulSoup(requests.get(url).text, "lxml")
        return soup.select_one(".tools-step > a.step-topic").attrs["href"].split("/")[-1]

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(
                r"https://www.kuaikanmanhua.com/((web/topic/\d*/?)|(webs/comic-next/\d+))", url
            )
        )


def get_class() -> type[CommonSource]:
    return KuaiKanManhuaSource



from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

from .ops import OutputProcessContainer, ProcessConfig, ProcessContainer

try:
    from PIL import Image, ImageFile

    ImageFile.LOAD_TRUNCATED_IMAGES = True

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False

    if not TYPE_CHECKING:

        class Image:
            class Image:
                pass


class ProcessOps(str, Enum):
    ROTATE_DOUBLE_PAGES = "rotate_double_pages"
    """If page width is greater than its height, rotate it 90 degrees."""

    SPLIT_DOUBLE_PAGES = "split_double_pages"
    """If page width is greater than its height, split it in half into two images."""

    NO_POSTPROCESSING = "none"
    """Disable image processing entirely."""

    TRIM_BORDERS = "trim_borders"
    """Conservatively remove borders in images."""

    RESIZE = "resize"
    """Resize images to a maximum width and height."""

    WEBP_TO_PNG = "webp_to_png"
    """Convert any WEBP images to PNG."""


"""A list of operations that can be chained."""
IN_MEM_PROCESS_OPS = {
    ProcessOps.ROTATE_DOUBLE_PAGES,
    ProcessOps.SPLIT_DOUBLE_PAGES,
    ProcessOps.TRIM_BORDERS,
    ProcessOps.RESIZE,
}

"""A list of operations that must be run last."""
OUTPUT_PROCESS_OPS = {
    ProcessOps.WEBP_TO_PNG,
}


class ProcessOptionMismatchError(Exception):
    """Raised when an option is not valid for a given operation or
    there are missing options."""


class Processor(ProcessContainer):
    """
    A class for processing images.

    :param `image_path`: The path to the image to process
    :raises `ImportError`: If Pillow is not installed
    :raises `ProcessOptionMismatchError`: If an option is not valid for
    a given operation or there are missing options
    """

    def __init__(self, image_path: Path | str, config: ProcessConfig | None = None) -> None:
        if not HAS_PILLOW:
            raise ImportError("Pillow was not found and is needed for processing. Is it installed?")

        super().__init__(config)
        self.image_path = Path(image_path)
        self._image = Image.open(self.image_path)
        self.is_modified = False

        # WARN: dangerous if there are multiple types of operations
        # that would add new image files to be written
        self.new_images: list[Image.Image] = []

    @property
    def image(self) -> Image.Image:
        """
        The image to process. This is a `PIL.Image.Image` object.

        :setter: Set the image to process
        :returns: The image to process
        """
        return self._image

    @image.setter
    def image(self, image: Image.Image) -> None:
        self._image = image
        self.is_modified = True

    def write(
        self, filename: Path | str | None = None, output_process_op: ProcessOps | None = None
    ) -> None:
        """Save the processed image(s) manually"""
        filename = Path(filename or self.image_path)

        writer = OutputProcessContainer(self.image, filename)
        getattr(writer, output_process_op or "default")()  # this is so janky

        for image in self.new_images:
            # increment by one "a" each time
            filename = filename.with_stem(filename.stem + "a")
            writer = OutputProcessContainer(image, filename)
            getattr(writer, output_process_op or "default")()

    def process(self, operations: list[ProcessOps], filename: Path | str | None = None) -> None:
        """
        Perform the operations in `operations` on the image in sequence
        and save it to disk. If `filename` is not None, it will be saved
        with that filename instead. If "none" is in `operations`, no
        post-processing will be done.

        :param operations: A list of operations to perform on the image
        :param filename: The filename to save the image as
        :raises NotImplementedError: If an operation is not implemented
        :raises OSError: If there is an error in saving the image
        :raises ProcessOptionMismatchError: If an option is not valid for a
        given operation or there are missing options
        """
        if ProcessOps.NO_POSTPROCESSING in operations:
            return

        # TODO: move all the checks together
        # there are some in ProcessConfig rn
        resize_op_valid = bool(self.config.output_profile or self.config.target_size) ^ bool(
            ProcessOps.RESIZE in operations
        )
        if resize_op_valid:
            # if any of the following is true:
            # - target_size is set and resize is not
            # - profile is set and resize is not
            # - resize is set and neither target_size nor profile is set
            # testing if only one of them is set is done in the resize op itself
            raise ProcessOptionMismatchError("resize must be used with target_size or profile")

        output_op: ProcessOps | None = None

        for func in operations:
            if func in OUTPUT_PROCESS_OPS:
                if output_op:
                    # only one output operation can be run
                    raise ProcessOptionMismatchError(
                        "Only one output operation can be run."
                        f"{output_op} and {func} were both found."
                    )
                output_op = func
                continue

            try:
                images: tuple[Image.Image, ...] | Image.Image | None = getattr(self, func)(
                    self.image
                )

                if images is None:
                    continue

                if isinstance(images, Image.Image):
                    images = (images,)

                self.image = images[0]

                if len(images) > 1:
                    self.new_images.extend(images[1:])

                self.is_modified = True
            except AttributeError as err:
                raise NotImplementedError(
                    f"{func} is not a valid post-processing function."
                ) from err
            except OSError as err:
                raise OSError(f"Error in {self.image_path}") from err

        if self.is_modified or output_op:
            # only write to disk if something has actually changed
            self.write(filename, output_op)



"""
Source file for comixextra.com
"""

import re

from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from ..request_utils import requests
from .common_source import CommonSource


class ComixExtraSource(CommonSource):
    name = "Comix Extra"
    domains = ["https://comixextra.com"]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)  # we make a GET request here which is not ideal
        self.url = f"https://comixextra.com/comic/{self.id}"
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        title: str = soup.select_one(".mobile-hide span.title-1").text.strip()

        authors: list[str] = []
        genres: list[str] = []
        for item in soup.select_one("dl.movie-dl").children:
            if item.name == "dt":
                if item.text == "Author:":
                    authors = item.find_next_sibling("dd").text.strip().split(",")
                elif item.text == "Genres:":
                    genres = [g.text.strip() for g in item.find_next_sibling("dd").select("a")]

        cover_art: str = soup.select_one(".movie-image img")["src"]

        description = soup.select_one("#film-content").text.strip()

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        return list(
            reversed(
                [
                    BaseChapter(c.text, f"{c['href']}/full")
                    for c in soup.select(".episode-list #list a")
                ]
            )
        )

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        text = requests.get(chapter.url).text

        return [i["src"] for i in BeautifulSoup(text, "lxml").select(".chapter-container img")]

    @classmethod
    def url_to_id(cls, url: str) -> str:
        items = list(filter(None, url.split("/")))

        # e.g., for https://comixextra.com/comic/nottingham and
        # https://comixextra.com/nottingham/issue-12
        # we want 'nottingham'

        if "comic" in items:
            return items[items.index("comic") + 1]
        return items[items.index("comixextra.com") + 1]

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://comixextra.com/.*", url))


def get_class() -> type[CommonSource]:
    return ComixExtraSource



"""
Source file for mangasee123.com
"""
# pylint: disable=invalid-name

import json
import re

import feedparser
import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class MangaSeeSource(CommonSource):
    name = "MangaSee"
    domains = ["https://mangasee123.com"]
    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:77.0) Gecko/20100101 Firefox/77.0"

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        # all of the metadata is available in the <script type="application/ld+json"> element
        # so are the chapters
        soup = BeautifulSoup(self._get_scripts(), "lxml")
        metadata_json: dict = json.loads(
            soup.find("script", type="application/ld+json").next_element
        )["mainEntity"]

        title: str = metadata_json["name"]

        authors_html = soup.select('a[href^="/search/?author"]')
        authors: list[str] = [a.next_element for a in authors_html]

        genres: list[str] = metadata_json["genre"]

        description_html = soup.select_one("div.top-5.Content")
        description = str(description_html.next_element).strip()

        cover_art = soup.select_one('div[style="padding-right:0px;"] img.bottom-5')["src"]

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        feed = feedparser.parse(f"https://mangasee123.com/rss/{self.id}.xml", agent=self.USER_AGENT)

        chapters = []
        for c in feed["entries"]:
            chapter_title = str(c["title"]).lstrip(self.metadata.title).strip()
            chapters.append(BaseChapter(chapter_title, c["link"]))

        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        full_js = str(
            soup.find("script", type="application/ld+json").find_next("script").next_element
        )

        index_start = full_js.index("vm.CurChapter =") + len("vm.CurChapter =")
        index_end = full_js.index(";", index_start)
        chapter_details: dict = json.loads(full_js[index_start:index_end])

        index_start = full_js.index("vm.CurPathName =") + len("vm.CurPathName =")
        index_end = full_js.index(";", index_start)

        domain: str = json.loads(full_js[index_start:index_end])
        num_pages = int(chapter_details["Page"])
        chapter_number = chapter.title.split()[-1]
        # pad to 4 excluding decimal places
        pad_zero = max(4 - len(str(int(float(chapter_number)))), 0)

        image_list = [
            f"https://{domain}/manga/{self.id}/{'0' * pad_zero}{chapter_number}-{i:03}.png"
            for i in range(1, num_pages + 1)
        ]
        return image_list

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://mangasee123.com/manga/.*", url))

    @classmethod
    def url_to_id(cls, url: str) -> str:
        # converts page url to id
        # e.g. https://mangasee123.com/manga/Kaguya-Wants-To-Be-Confessed-To
        # to Kaguya-Wants-To-Be-Confessed-To
        *_, last_item = filter(None, url.split("/"))
        return last_item


def get_class() -> type[CommonSource]:
    return MangaSeeSource



"""
Source file for webtoons.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class WebtoonsSource(CommonSource):
    name = "Webtoons"
    domains = ["https://webtoons.com"]
    headers = {"Referer": "https://webtoons.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self._soup: BeautifulSoup | None = None

        title_no_index = url.index("?title_no=") + len("?title_no=")
        title_no_end_index: int | None = url.find("&", title_no_index)
        if title_no_end_index == -1:  # if '&' not found in url
            title_no_end_index = None

        self._title_no = int(url[title_no_index:title_no_end_index])

        self._title_path = "/".join(url.split("/")[3:6])

    def fetch_metadata(self) -> BaseMetadata:
        page = self._get_desktop_soup()
        title = page.select_one('meta[property="og:title"]')["content"]
        authors: list[str] = [
            s.strip()
            for s in page.select_one('meta[property="com-linewebtoon:webtoon:author"]')[
                "content"
            ].split("/")
        ]
        description: str = page.select_one("#content .summary").text
        cover_art_el = page.select_one("#content .detail_body.banner")

        if cover_art_el is None:
            # canvas challenge webtoons can have multiple genres
            # and have different covers
            cover_art_el = page.select_one("#content .detail_header.challenge img")

            cover_art = cover_art_el["src"]
        else:
            art_start_idx = cover_art_el["style"].find("url(") + len("url(")
            art_end_idx = cover_art_el["style"].find(")", art_start_idx)

            cover_art = cover_art_el["style"][art_start_idx:art_end_idx]

        genres_els = list(page.select(".info .genre"))

        for el in genres_els:
            span = el.find_all("span")
            for s in span:
                s.replace_with("")

        genres = [el.text for el in page.select(".info .genre")]

        return BaseMetadata(
            title,
            authors,
            f"https://www.webtoons.com/{self._title_path}/list?title_no={self._title_no}",
            genres,
            description,
            cover_art,
        )

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = self._get_soup()
        titles = [str(e.next_element) for e in soup.select("p.sub_title > span.ellipsis")]

        links: list[str] = [e["href"] for e in soup.select('a[class^="NPI=a:list"]')]

        chapters = [BaseChapter(t, u) for t, u in zip(titles, links)]
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        images: list[str] = []
        for c in soup.select("div#_imageList > img"):
            images.append(c["data-url"])
        return images

    def _get_soup(self) -> BeautifulSoup:
        if self._soup:
            return self._soup

        # mobile serves all chapters in one page
        mobile_url = f"https://m.webtoons.com/{self._title_path}/list?title_no={self._title_no}"

        self._soup = BeautifulSoup(
            requests.get(mobile_url, headers=self.headers).text,
            "lxml",
        )
        return self._soup

    def _get_desktop_soup(self) -> BeautifulSoup:
        desktop_url = f"https://www.webtoons.com/{self._title_path}/list?title_no={self._title_no}"
        return BeautifulSoup(requests.get(desktop_url, headers=self.headers).text, "lxml")

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(r"https://www.webtoons.com/.*/list\?title_no=.*", url)
            or re.match(r"https://m.webtoons.com/.*/list\?title_no=.*", url)
            or re.match(r"https://www.webtoons.com/.*/viewer\?title_no=.*", url)
        )


def get_class() -> type[CommonSource]:
    return WebtoonsSource



from .api import (
    ConvertFormats,
    convert,
    convert_progress,
    download,
    download_progress,
    load,
    process,
    process_progress,
    query,
    save_metadata,
)
from .base import BaseChapter, BaseMetadata
from .comic import BaseComic
from .io import MD_METADATA_FILE
from .processor import (
    ProcessConfig,
    ProcessOps,
    ProcessOptionMismatchError,
    Processor,
)
from .processor.profiles import SupportedProfiles, all_profiles

__version__ = (1, 10, 2)
__version_str__ = ".".join(map(str, __version__))



"""
Source file for readcomiconline.li
"""
# pylint: disable=invalid-name

import binascii
import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from ..errors import NoImagesFoundError
from .common_source import CommonSource


class ReadComicOnlineSource(CommonSource):
    name = "ReadComicOnline"
    domains = ["https://readcomiconline.li"]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(
            requests.get(f"https://readcomiconline.li/Comic/{self.id}").text,
            "lxml",
        )

        title = str(soup.select_one("h3").text)
        author = [
            str(
                # this site uses "Various" if there's more than one author
                soup.select_one("a[href^='/Writer']").text
            )
        ]
        genres: list[str] = [str(e.text) for e in soup.select("a[href^='/Genre']")]
        description_maybe = soup.select_one("p[style='text-align: justify;']")

        if description_maybe is not None:
            for br in description_maybe.find_all("br"):
                br.replace_with("\n")
            description = str(description_maybe.text)
        else:
            description = ""
        cover = self.domains[0] + str(soup.find("link")["href"])

        return BaseMetadata(title, author, self.url, genres, description, cover)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(
            requests.get(f"https://readcomiconline.li/Comic/{self.id}").text,
            "lxml",
        )

        chapters: list[BaseChapter] = [
            BaseChapter(e.text.strip(), f'{self.domains[0]}{e["href"]}')
            for e in soup.select("ul.list > li > div > a")
        ]
        return list(reversed(chapters))

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        text = requests.get(chapter.url).text

        images: list[str] = []
        start = 0
        while (index := text.find("lstImages.push(", start)) != -1:
            s_index = index + len("lstImages.push(") + 1
            e_index = text.find(");", s_index) - 1  # could be single or double quotes
            images.append(self.beau(text[s_index:e_index]))
            start = e_index
        if not images:
            raise NoImagesFoundError(
                "No images found in the current chapter. Try visiting the URL in your browser "
                "and solving any CAPTCHAs before trying again."
            )
        return images

    @classmethod
    def url_to_id(cls, url: str) -> str:
        segments = url.split("/")
        for i, s in enumerate(segments):
            if s == "Comic":
                return segments[i + 1]
        raise ValueError("Invalid comic URL")

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://readcomiconline.li/Comic/.*", url))

    @staticmethod
    def beau(url: str) -> str:
        """
        Deobfuscator copied from
        https://github.com/mikf/gallery-dl/blob/master/gallery_dl/extractor/readcomiconline.py

        i had to relicense to GPLv2 for this >:(
        """
        url = url.replace("_x236", "d").replace("_x945", "g")

        if url.startswith("https"):
            return url

        url, sep, rest = url.partition("?")
        containsS0 = "=s0" in url
        url = url[: -3 if containsS0 else -6]
        url = url[4:22] + url[25:]
        url = url[0:-6] + url[-2:]
        url = binascii.a2b_base64(url).decode()
        url = url[0:13] + url[17:]
        url = url[0:-2] + ("=s0" if containsS0 else "=s1600")
        return f"https://2.bp.blogspot.com/{url}{sep}{rest}"


def get_class() -> type[CommonSource]:
    return ReadComicOnlineSource



# pylint: disable=invalid-name

import shutil
from pathlib import Path
from typing import Iterator

import comicon

from . import io, sources
from .comic import BaseComic
from .convert_utils import ConvertFormats, convert_one
from .errors import ImageDownloadError
from .processor import ProcessConfig, ProcessOps, Processor


def query(url: str) -> BaseComic:
    """
    Attempt to query for a comic given a URL.
    :param `url`: An internet URL to search for
    :raises `ValueError` if the source is not found.
    """
    adapter = sources.get_class_for(url)(url)
    return BaseComic(adapter.metadata, adapter.chapters)


def load(path: Path | str) -> BaseComic:
    """
    Load a mandown-created comic from the file system.

    :param `path`: A folder where mandown has created a comic
    :returns A comic with metadata and chapter data of that folder

    :raises FileNotFoundError if `md-metadata.json` cannot be found
    :raises IOError if the path does not exist
    """
    return io.read_comic(path)


def save_metadata(comic: BaseComic, path: Path | str) -> None:
    """
    Save the metadata from the comic to `<path>/md-metadata.json`.

    :param `comic`: A comic with metadata to save
    :param `path`: A folder to save the metadata to
    """
    io.save_comic(comic, path)


def init_parse_comic(
    path: Path | str, donor_comic: BaseComic | None = None, download_cover: bool = False
) -> BaseComic:
    """
    Open a comic from a folder path, either via `md-metadata.json` or
    if that fails, parse the comic structure and create an `md-metadata.json`

    :param `path`: A folder containing `md-metadata.json` or a comic structure
    :param `data`: A comic to fill metadata from if no metadata is found
    :param `download_cover`: If `True` and `source_url` is set, download
    the cover image if no metadata is found
    :returns A comic with metadata and chapter data of that folder
    :raises `AttributeError` if the source URL is not set and `download_cover` is `True`
    """
    if not donor_comic and download_cover:
        raise AttributeError("Cannot download cover without donor comic")

    try:
        comic = io.read_comic(path)
    except FileNotFoundError:
        comic = io.parse_comic(path, donor_comic)
        io.save_comic(comic, path)

        if download_cover and comic.metadata.cover_art:
            next(
                io.download_images(
                    [comic.metadata.cover_art],
                    path,
                    filestems=["cover"],
                    headers=comic.source.headers,
                )
            )
    return comic


def convert_progress(
    comic_path: Path | str,
    to: ConvertFormats,
    dest_folder: Path | str | None = None,
    remove_after: bool = False,
    split_by_chapters: bool = False,
) -> Iterator[str | int]:
    """
    Convert the comic located at `folder_path` to `convert_to`
    and put it in `dest_folder` (defaults to workdir).

    :param `comic_path`: The path to the comic to convert (may be in Mandown
    folder form or any of the `mandown.ConvertFormats` such as EPUB)
    :param `convert_to`: The format to convert to
    :param `dest_folder`: A folder to put the converted comic in
    :param `remove_after`: If `True`, delete the original file/folder after conversion
    :param `split_by_chapters`: Only applies to Mandown-created comics. If `True`,
    output a comic file per chapter. Existing comic files will not be overwritten.

    :returns An `Iterator` representing a progress bar. The first iteration returns
    the remaining number of iterations. If converting between file formats, an
    iteration after the first number of iterations ends will return the remaining
    number of the second number of iterations.
    """
    comic_path = Path(comic_path)
    if to == ConvertFormats.NONE:
        return

    # default to working directory
    dest_folder = Path(dest_folder or ".").resolve()

    if comic_path.is_dir():
        # it's a mandown comic, convert it to CIR
        comic = load(comic_path)

        # find cover
        cover: str | None = None
        for item in comic_path.iterdir():
            if item.name.startswith("cover"):
                cover = item.name
                break

        if split_by_chapters:
            comicon_comics = [
                comicon.Comic(
                    comicon.Metadata(
                        title=f"{comic.metadata.title} - {chap.title}",
                        authors=comic.metadata.authors,
                        description=comic.metadata.description,
                        genres=comic.metadata.genres,
                        cover_path_rel=cover,
                    ),
                    [comicon.Chapter(chap.title, chap.slug)],
                )
                for chap in comic.chapters
            ]

            yield len(comicon_comics)
            existing_filenames = {file.name for file in dest_folder.iterdir() if file.is_file()}
            for comicomic in comicon_comics:
                yield comicomic.metadata.title

                # do not overwrite existing cache
                if f"{comicomic.metadata.title_slug}.{to.value}" in existing_filenames:
                    continue
                for _ in convert_one(comicomic, comic_path, to, dest_folder):
                    ...

        else:
            comicon_comic = comicon.Comic(
                comicon.Metadata(
                    title=comic.metadata.title,
                    authors=comic.metadata.authors,
                    description=comic.metadata.description,
                    genres=comic.metadata.genres,
                    cover_path_rel=cover,
                ),
                [comicon.Chapter(chap.title, chap.slug) for chap in comic.chapters],
            )

            yield from convert_one(comicon_comic, comic_path, to, dest_folder)

    else:
        # it's a file, no conversion needed, let comicon do its inferencing
        yield from comicon.convert_progress(
            comic_path, dest_folder / f"{comic_path.stem}.{to.value}"
        )

    if remove_after:
        shutil.rmtree(comic_path)


def convert(
    comic_path: Path | str,
    to: ConvertFormats,
    dest_folder: Path | str | None = None,
    remove_after: bool = False,
) -> None:
    """
    Convert the comic located at `folder_path` to `convert_to`
    and put it in `dest_folder` (defaults to workdir).

    :param `comic_path`: The path to the comic to convert (may be in Mandown
    folder form or any of the `mandown.ConvertFormats` such as EPUB)
    :param `convert_to`: The format to convert to
    :param `dest_folder`: A folder to put the converted comic in
    :param `remove_after`: If `True`, delete the original file/folder after conversion
    """
    for _ in convert_progress(comic_path, to, dest_folder, remove_after):
        pass


def process_progress(
    comic_path: Path | str, ops: list[ProcessOps], config: ProcessConfig | None = None
) -> Iterator[str]:
    """
    Process the comic in `comic_path` with `ops` in the order provided.

    :param `comic_path`: A folder containing a image folders to process
    :param `ops`: A list of operations to perform on each image
    :param `config`: Options for processing operations
    :returns An `Iterator` representing a progress bar up to the number of images in the comic.
    """
    data = io.discover_local_images(comic_path)
    for _, images in data.items():
        for i in images:
            Processor(i, config).process(ops)
        yield "Processing"


def process(
    comic_path: Path | str, ops: list[ProcessOps], config: ProcessConfig | None = None
) -> None:
    """
    Process the comic in `comic_path` with `ops` in the order provided.

    :param `comic_path`: A folder containing a image folders to process
    :param `ops`: A list of operations to perform on each image
    :param `config`: Options for processing operations
    """
    for _ in process_progress(comic_path, ops, config):
        pass


def download_progress(
    comic: BaseComic | str,
    path: Path | str = ".",
    *,
    start: int | None = None,
    end: int | None = None,
    threads: int = 4,
    only_download_missing: bool = True,
    raise_on_failed_download: bool = True,
) -> Iterator[str]:
    """
    Download comic or comic URL `comic` to `path` using `threads` threads.

    :param `comic`: A comic or URL to download
    :param `path`: A folder to download the comic to
    :param `start`: The first chapter to download (zero-indexed, inclusive)
    :param `end`: The last chapter to download (zero-indexed, exclusive)
    :param `threads`: The number of threads to use
    :param `only_download_missing`: If `True`, do not download
    images already in the destination path

    :returns An `Iterator` representing a progress bar up to the number of chapters in the comic.
    """
    path = Path(path)

    # make var comic a BaseComic
    if isinstance(comic, str):
        comic = query(comic)

    full_path = path / comic.metadata.title_slug
    full_path.mkdir(exist_ok=True)

    # save metadata json
    comic.set_chapter_range(start=start, end=end)
    io.save_comic(comic, full_path)

    # cover
    if comic.metadata.cover_art:
        for _ in io.download_images(
            [comic.metadata.cover_art],
            full_path,
            filestems=["cover"],
            headers=comic.source.headers,
        ):
            pass

    # for each chapter
    for chap in comic.chapters[start:end]:
        yield chap.title
        image_urls = comic.get_chapter_image_urls(chap)
        chapter_path = full_path / chap.slug
        chapter_path.mkdir(exist_ok=True)

        # expect that they're named by numbers only
        skip_images: set[int] = set()
        if only_download_missing:
            for file in chapter_path.iterdir():
                if file.stem == file.stem.rjust(io.NUM_LEFT_PAD_DIGITS, "0"):
                    try:
                        skip_images.add(int(file.stem))
                    except ValueError:
                        # expected if not an image file
                        pass

        if not image_urls or len(skip_images) == len(image_urls):
            # move to next chapter if there's nothing to download for this one
            continue

        # name them 00001.png, 00002.png, etc
        # skipping ones that already exist
        processed_image_urls, filestems = zip(
            *(
                (link, str(i).rjust(io.NUM_LEFT_PAD_DIGITS, "0"))
                for i, link in enumerate(image_urls, start=1)
                if i not in skip_images
            ),
            strict=False,
        )

        chapter_path = full_path / chap.slug

        for _ in io.download_images(
            processed_image_urls,
            chapter_path,
            headers=comic.source.headers,
            filestems=filestems,
            threads=threads,
        ):
            pass

        # check if every image was downloaded
        if count := len([f for f in chapter_path.iterdir() if f.is_file()]) != len(
            processed_image_urls
        ):
            if raise_on_failed_download:
                raise ImageDownloadError(
                    f"Failed to download {len(processed_image_urls) - count} images"
                )


def download(
    comic: BaseComic | str,
    path: Path | str = ".",
    *,
    start: int | None = None,
    end: int | None = None,
    threads: int = 4,
    only_download_missing: bool = True,
    raise_on_failed_download: bool = True,
) -> None:
    """
    Download comic or comic URL `comic` to `path` using `threads` threads.

    :param `comic`: A comic or URL to download
    :param `path`: A folder to download the comic to
    :param `start`: The first chapter to download (one-indexed, inclusive)
    :param `end`: The last chapter to download (one-indexed, inclusive)
    :param `threads`: The number of threads to use
    :param `only_download_missing`: If `True`, do not download images
    already in the destination path
    """
    for _ in download_progress(
        comic,
        path,
        start=start,
        end=end,
        threads=threads,
        only_download_missing=only_download_missing,
    ):
        pass



"""
Source file for mangadex.org
"""
# pylint: disable=invalid-name

import re
import time

import requests
from bs4 import BeautifulSoup
from slugify import slugify

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class MangaDexSource(CommonSource):
    name = "MangaDex"
    domains = ["https://mangadex.org"]

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self._soup: BeautifulSoup | None = None
        self.lang_code = ""

        # https://api.mangadex.org/manga/de4b3c43-5243-4399-9fc3-68a3c0747138
        self.id = self.url.split("/")[4]
        if self.url.startswith("https://mangadex.org/chapter"):
            r: dict = self._get(f"https://api.mangadex.org/chapter/{self.id}").json()["data"][
                "relationships"
            ]
            self.id: str = next(filter(lambda i: i["type"] == "manga", r))["id"]  # type: ignore

    def fetch_metadata(self) -> BaseMetadata:
        # TODO: support non-English downloads
        r = self._get(
            f"https://api.mangadex.org/manga/{self.id}"
            "?includes[]=author&includes[]=cover_art&includes[]=artist"
        ).json()

        metadata: dict = r["data"]

        # use english if possible, otherwise use the first language that appears
        self.lang_code = (
            "en"
            if "en" in metadata["attributes"]["title"]
            else next(iter(metadata["attributes"]["title"]))
        )
        title: str = metadata["attributes"]["title"][self.lang_code]

        if metadata["attributes"]["description"]:
            description: str = metadata["attributes"]["description"][self.lang_code]

            # strip trailing spaces on each line
            description = "\n".join(s.strip() for s in description.split("\n"))
        else:
            description = ""

        authors: set[str] = set()
        cover_art = ""
        for d in metadata["relationships"]:
            if d["type"] == "author" or d["type"] == "artist":
                authors.add(d["attributes"]["name"])
            elif d["type"] == "cover_art":
                # pylint: disable=line-too-long
                cover_art = (
                    "https://uploads.mangadex.org/covers/"
                    f"{self.id}/{d['attributes']['fileName']}"
                )

        genres: list[str] = []
        for d in metadata["attributes"]["tags"]:
            tag = d["attributes"]["group"]
            if tag == "genre" or tag == "theme":
                genres.append(d["attributes"]["name"][self.lang_code])

        return BaseMetadata(
            title,
            list(authors),
            f"https://mangadex.org/title/{self.id}",
            genres,
            description,
            cover_art,
        )

    def fetch_chapter_list(self) -> list[BaseChapter]:
        # for some reason *sometimes* it goes all name/service not found
        r = self._get(
            f"https://api.mangadex.org/manga/{self.id}/"
            f"feed?limit=500&translatedLanguage[]={self.lang_code}"
            "&order[volume]=asc&order[chapter]=asc"
        ).json()

        chapters: list[BaseChapter] = []
        for i, c in enumerate(r["data"]):
            chapter_title: str = c["attributes"]["title"] or f"Chapter {c['attributes']['chapter']}"
            chapter_slug: str = f"{i}-{slugify(chapter_title)}"
            chapters.append(
                BaseChapter(
                    chapter_title,
                    f"https://mangadex.org/chapter/{c['id']}",
                    chapter_slug,
                )
            )
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        *_, chapter_id = chapter.url.split("/")
        r = self._get(f"https://api.mangadex.org/at-home/server/{chapter_id}").json()
        base_url = r["baseUrl"]
        chapter_hash = r["chapter"]["hash"]

        images: list[str] = []
        for p in r["chapter"]["data"]:
            images.append(f"{base_url}/data/{chapter_hash}/{p}")
        return images

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(
            re.match(r"https://mangadex.org/title/.*", url)
            or re.match(r"https://mangadex.org/chapter/.*", url)
        )

    @staticmethod
    def _get(url: str) -> requests.Response:
        """
        A wrapper of requests.get for MangaDex with
        rudimentary rate-limit processing
        """
        for _ in range(3):
            r = requests.get(url, timeout=5)
            if r.status_code == 200:
                break
            elif r.status_code == 404:
                raise RuntimeError(
                    "This chapter is not downloadable from MangaDex. If you "
                    "believe this to be an error, please open a GitHub issue."
                )
            time.sleep(1)
        else:
            raise RuntimeError("MangaDex is probably rate-limiting us, try again later?")
        return r


def get_class() -> type[CommonSource]:
    return MangaDexSource



from dataclasses import dataclass
from typing import Literal


@dataclass(slots=True, frozen=True)
class OutputProfile:
    id: str
    name: str
    size: tuple[int, int]


SupportedProfiles = Literal[
    "kindle",
    "kintouch",
    "paper",
    "voyage",
    "oasis2",
    "kobotouch",
    "glo",
    "glohd",
    "aura",
    "aurahd",
    "aurah2o",
    "aurah2o2",
    "auraone",
    "clarahd",
    "librah2o",
    "nia",
    "clara2e",
    "libra2",
    "sage",
]
_ProfileType = tuple[str, tuple[int, int]]

__all_profiles_data: dict[SupportedProfiles, _ProfileType] = {
    # Kindles
    "kindle": ("Kindle", (600, 800)),
    "kintouch": ("Kindle Touch", (600, 800)),
    "paper": ("Kindle Paperwhite 1/2", (758, 1024)),
    "voyage": ("Kindle Voyage, Oasis, Paperwhite 3/4", (1072, 1448)),
    "oasis2": ("Kindle Oasis 2/3", (1264, 1680)),
    # Kobos
    "kobotouch": ("Kobo Mini/Touch", (600, 800)),
    "glo": ("Kobo Glo", (768, 1024)),
    "glohd": ("Kobo Glo HD", (1072, 1448)),
    "aura": ("Kobo Aura", (758, 1024)),
    "aurahd": ("Kobo Aura HD", (1080, 1440)),
    "aurah2o": ("Kobo Aura H2O", (1080, 1430)),
    "aurah2o2": ("Kobo Aura H2O 2", (1080, 1430)),
    "auraone": ("Kobo Aura One", (1404, 1872)),
    "clarahd": ("Kobo Clara HD", (1072, 1448)),
    "librah2o": ("Kobo Libra H2O", (1264, 1680)),
    "nia": ("Kobo Nia", (758, 1024)),
    "clara2e": ("Kobo Clara 2E", (1072, 1448)),
    "libra2": ("Kobo Libra 2", (1264, 1680)),
    "sage": ("Kobo Sage", (1440, 1920)),
}

all_profiles = {id: OutputProfile(id, *profile) for id, profile in __all_profiles_data.items()}



from ..request_utils import requests
from .base_source import BaseSource


class CommonSource(BaseSource):
    _scripts: str | None = None

    def _get_scripts(self) -> str:
        """
        Legacy method for fetching the HTML of `self.url`.
        """
        if self._scripts:
            return self._scripts

        self._scripts = requests.get(self.url).text or ""
        return self._scripts



from dataclasses import dataclass

from comicon import SLUGIFY_ARGS
from slugify import slugify


@dataclass(slots=True)
class BaseMetadata:
    """
    Metadata for a comic.

    :param `title`: The title of the comic
    :param `authors`: A list of authors of the comic
    :param `url`: The URL of the comic
    :param `genres`: A list of genres of the comic
    :param `description`: A description of the comic
    :param `cover_art`: The URL of the cover art
    :param `title_slug`: The slug of the title
    """

    title: str
    authors: list[str]
    url: str
    genres: list[str]
    description: str
    cover_art: str
    title_slug: str = ""

    def __post_init__(self) -> None:
        self.title_slug = slugify(self.title, **SLUGIFY_ARGS)

    def asdict(self) -> dict[str, str | list[str]]:
        """
        Return a dictionary representation of the metadata.
        """

        return {
            "title": self.title,
            "authors": self.authors,
            "url": self.url,
            "genres": self.genres,
            "description": self.description,
            "cover_art": self.cover_art,
        }


@dataclass(slots=True)
class BaseChapter:
    """
    A chapter of a comic.

    :param `title`: The title of the chapter
    :param `url`: The URL of the chapter
    :param `slug`: The slug of the chapter
    """

    title: str
    url: str
    slug: str = ""

    def __post_init__(self) -> None:
        if not self.slug:
            self.slug = slugify(self.title, **SLUGIFY_ARGS)

    def asdict(self) -> dict:
        """
        Return a dictionary representation of the chapter.
        """
        return {
            "title": self.title,
            "url": self.url,
            "slug": self.slug,
        }



"""
Source file for blogtruyenmoi.com
"""
# pylint: disable=invalid-name

import re

import requests
from bs4 import BeautifulSoup

from ..base import BaseChapter, BaseMetadata
from .common_source import CommonSource


class BlogTruyenMoiSource(CommonSource):
    name = "Blog Truyen"
    domains = ["https://blogtruyenmoi.com"]
    headers = {"Referer": "https://blogtruyenmoi.com/"}

    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.id = self.url_to_id(url)  # we make a GET request here which is not ideal
        self.url = f"https://blogtruyenmoi.com/{self.id}"
        self._scripts: str | None = None

    def fetch_metadata(self) -> BaseMetadata:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        title: str = (
            soup.select_one("div#breadcrumbs > span:nth-child(2)").text.split(" > ")[-1].strip()
        )
        authors: list[str] = [soup.select_one('div.description > p > a[href^="/tac-gia"]').text]
        genres: list[str] = [string.text.strip() for string in soup.select("span.category > a")]
        cover_art: str = soup.select_one(".thumbnail > img")["src"]

        description_list: list[str] = []
        for child in soup.select_one("div.detail > div.content").children:
            if child.name and child.name.lower() == "br":
                # br tags usually indicate the end of the description
                break

            stripped = child.text.strip()
            if not stripped:
                break
            description_list.append(stripped)
        description = "\n".join(description_list)

        return BaseMetadata(title, authors, self.url, genres, description, cover_art)

    def fetch_chapter_list(self) -> list[BaseChapter]:
        soup = BeautifulSoup(self._get_scripts(), "lxml")

        chapters = [
            BaseChapter(c.text.strip(), f"https://blogtruyenmoi.com/{c['href']}")
            for c in soup.select("div#list-chapters > p > span > a")
        ]
        chapters.reverse()
        return chapters

    def fetch_chapter_image_list(self, chapter: BaseChapter) -> list[str]:
        soup = BeautifulSoup(requests.get(chapter.url).text, "lxml")
        return [el["src"] for el in soup.select("article#content > img")]

    @classmethod
    def url_to_id(cls, url: str) -> str:
        items = list(filter(None, url.split("/")))
        # e.g., for https://blogtruyenmoi.com/31844/doc-la-dungeon
        # we want '31844'
        # for https://blogtruyenmoi.com/c842985/doc-la-dungeon-chap-62
        # we want to crawl the page and find the link back to the ID

        try:
            return int(items[2])
        except ValueError:
            soup = BeautifulSoup(requests.get(url).text, "lxml")
            return (
                soup.select_one("header div.breadcrumbs > a:not([href='/'])")
                .attrs["href"]
                .split("/")[1]
            )

    @staticmethod
    def check_url(url: str) -> bool:
        return bool(re.match(r"https://blogtruyenmoi.com/(\d*(/.*)?)|(c.\d*/.*)", url))


def get_class() -> type[CommonSource]:
    return BlogTruyenMoiSource
